/* $Id: c_pilot.c,v 1.9 2004/02/23 04:36:00 mjhammel Exp $ */

/*========================================================================
 *  Copyright (c) Michael J. Hammel 1995-2001
 *========================================================================
 *              FILE NAME: pilotaddr.c
 *            DESCRIPTION: Parser routines for pilot address files
 *      DEFINED CONSTANTS: 
 *       TYPE DEFINITIONS: 
 *      MACRO DEFINITIONS: 
 *       GLOBAL VARIABLES: 
 *       PUBLIC FUNCTIONS: XPPilotParseAddr, XPPilotParseTodo, XPWriteTodos
 *                       : XPWriteAddresses
 *                       : XPPilotParseAppointment
 *                       : XPPilotWriteAppointment
 *      PRIVATE FUNCTIONS: SortAddresses SortFields
 *                       : FreeAddressList FreeCategoryList
 *                       : GenerateCategoryList
 *  SOFTWARE DEPENDENCIES: 
 *  HARDWARE DEPENDENCIES: 
 *                  NOTES: 
 *
 * SPECIAL CONSIDERATIONS:
 * Set your tabstops to 3 to make the code more readable.
 *========================================================================
 *
 * MODIFICATION HISTORY:
 * $Log: c_pilot.c,v $
 * Revision 1.9  2004/02/23 04:36:00  mjhammel
 * Calendar complete for beta 1.
 * Initial documentation updates for 3.6.
 * Fixed bug in FindText.c - accept button should be "gshell.search", not
 * "gshell.accept".
 *
 * Revision 1.8  2004/02/20 04:49:23  mjhammel
 * Added CUPS support for printing.
 * Fixed monthly display in appt's to only highlight date for current month.
 * Changed notebook layout in appt's.
 *
 * Revision 1.7  2004/02/18 06:07:42  mjhammel
 * Changed hilight for monthly page to occur in SetupMonthlyPage().
 * Changed ApptWrite routine to unlink file if link list is empty.
 *
 * Revision 1.6  2004/02/17 14:25:54  mjhammel
 * Added first pass at alarm handling for appointments.
 *
 * Revision 1.5  2004/02/11 03:59:19  mjhammel
 * Added pixmaps to monthly view of calendar.
 * Fixed today display in calendar to show all entries (but more fixes are
 * needed).
 * Added alarm handling to calendar.
 * Added new entry to calendar and made new and edited entries get saved
 * immediately.
 * Fixed much of the repeat handling in the calendar UI.
 *
 * Revision 1.4  2004/01/23 04:22:41  mjhammel
 * Added WriteAppt and FreeAppt to c_pilot.c
 * Added appointment related defines to XNotesplus.h
 * Added appointment prototypes to prototypes.h
 * Added PopulateCalendar() stub to appointments.c
 * Fixed bug in todos that caused crash when todo entry was deleted.
 *
 * Revision 1.3  2004/01/19 03:26:00  mjhammel
 * XPPilotParseAppointment() for c_pilot.c added and tested.
 *
 * Revision 1.2  2004/01/16 14:54:13  mjhammel
 * Adding Appointment file parsing to c_pilot.c.
 *
 * Revision 1.1.1.1  2002/01/10 18:37:51  mjhammel
 * Initial checkin
 *
 *
 *========================================================================*/
#define XP_CPADDR_C

/* === System Headers === */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <gtk/gtk.h> 

/* === Project Headers === */
#include "XNotesPlus.h"
#include "debug.h"
#include "XNpilot.h"

/* === external routines === */

/* === Public routine prototypes === */
void XPFreeApptList();
void XPPilotParseDBList();
void XPFreeMemoList();
int  XPPilotParseMemos();
void XPWriteMemos();
int  XPPilotParseAddr();
void XPWriteAddresses();
void XPAddAddress();
void XPFreeCategoryList();
void XPFreeAddressList();
void XPGenerateMemoCategoryList();
void XPWriteTodos();
int  XPPilotParseTodo();
void XPPilotWriteAppointment();
int  XPPilotParseAppointment();

/* === Private routine prototypes === */
static void FreeTodoList();
static void GenerateCategoryList();

static void PrintFields();
static void SortAddresses();
static void SortCategoryList();
static gboolean SortFields();

/* === Global Variables === */
PilotAddr_T					*Pilot_Address=NULL;
PilotCategory_T			*Pilot_Addr_Categories=NULL;
PilotTodo_T					*Pilot_Todo=NULL;
PilotCategory_T			*Pilot_Todo_Categories=NULL;
PilotMemo_T					*Pilot_Memo=NULL;
PilotCategory_T			*Pilot_Memo_Categories=NULL;
PilotAppt_T					*Pilot_Appt=NULL;
PilotCategory_T			*Pilot_Appt_Categories=NULL;


/* Ugliness from pilot-link:: reorganize fields to match visible appearence */
int realentry[19] =   { 0, 1, 13, 2, 
                        3, 4, 5, 6, 7, 
                        8, 9, 10, 11, 12,
                        14, 15, 16, 17,
                        18 };

char magicmark[0xffff];

/* Field names used in the local text file */
char *fieldnames[] = {
	"Category",
	"RecordID",
	"ListPhone",
	"Secret",
	"Last",
	"First",
	"Title",
	"Company",
	"<PHONE>",
	"<PHONE>",
	"<PHONE>",
	"<PHONE>",
	"<PHONE>",
	"Address",
	"City",
	"State",
	"ZipCode",
	"Country",
	"Custom1",
	"Custom2",
	"Custom3",
	"Custom4",
	"Note",
	NULL
};





/*========================================================================
 *	Name:			XPPilotParseAddr
 *	Prototype:	void XPPilotParseAddr(char *filename)
 *					
 *	Description:
 *	Parses the CSV output from pilot-addresses
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *		XP_E_FILE_EXIST		file to parse doesn't exist or is not a text file
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 * This is a complete rewrite from 3.3.0.  It is not compatible but works
 * much better.  However, much of this code is duplicated from the xn-d
 * directory because that directory has to be GPL while this code is
 * not.
 *		
 *========================================================================*/
int 
XPPilotParseAddr(
	char	*filename
)
{
#ifdef DEBUG
	char				*fname="XPPiloParseAddr()";
#endif

	struct stat		stat_buf;
	FILE				*fd;

	PilotAddr_T		*pentry;
	PilotAddr_T		*current_link;

	int				filelen;
	char				*file_text;
	char				*name, *value;
	int				cPtr, endPtr;
	int				count;

	DBGEnter();
	DBGPrintf(DBG_INFO,("pilot db to look for: %s\n", filename));

	/*
	 * Verify file exists.  Return if it doesn't.
	 */
	if ( stat(filename, &stat_buf) != 0 )
		return XP_E_FILE_EXIST;

	if ( ! ( stat_buf.st_mode & S_IFREG ) )
		return XP_E_FILE_EXIST;

	/*
	 * File is ok.  Open it.
	 */
	fd = fopen(filename, "r");

	/*
	 * Free up any old link lists.
	 */
	DBGPrintf(DBG_INFO,("Freeing old address and category lists.\n"));
	XPFreeCategoryList(&Pilot_Addr_Categories);
	XPFreeAddressList(&Pilot_Address);

	/*
	 * Allocate a new link for an address record.
	 */
	DBGPrintf(DBG_INFO,("Allocating new link for list.\n"));
	pentry = (PilotAddr_T *)XPSafeAlloc(sizeof(PilotAddr_T));
	memset(pentry,0,sizeof(PilotAddr_T));

	Pilot_Address = pentry;
	current_link = pentry;
	current_link->next = NULL;

	/* Read the file into memory - it's faster this way. */
	fseek(fd, 0, SEEK_END);
	filelen = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	file_text = (char*)malloc(filelen+1);
	if (file_text == NULL) 
	{
		/* This should never happen. */
		perror("malloc()");
		exit(1);
	}
   fread(file_text, filelen, 1, fd);

	/*
	 * Initialize pointers, counters used to parse input text.
	 */
	cPtr = (int)file_text;

	/* Skip the Magic string */
	name=strtok((char *)cPtr, "\n");
	cPtr = cPtr + strlen(name) + 1;


	/*
	 * Skip the category list - we can generate that later from the records.
	 */
	while( strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 )
	{
		name=strtok((char *)cPtr, "\n");
		GenerateCategoryList(name, &Pilot_Addr_Categories);

		cPtr = cPtr + strlen(name) + 1;
	}

	count = 0;

	/* 
	 * Parse the text from the input file.  The format of each entry in
	 * the input files must be in the following name/value formats:
	 *
	 * Category: <multi line of data>
	 * RecordID: <multi line of data>
	 * ListPhone: <multi line of data>
	 * Attribute: <multi line of data>
	 * Last: <multi line of data>
	 * First: <multi line of data>
	 * Title: <multi line of data>
	 * Company: <multi line of data>
	 * <PHONE>: <multi line of data>
	 * <PHONE>: <multi line of data>
	 * <PHONE>: <multi line of data>
	 * <PHONE>: <multi line of data>
	 * <PHONE>: <multi line of data>
	 * Address: <multi line of data>
	 * City: <multi line of data>
	 * State: <multi line of data>
	 * ZipCode: <multi line of data>
	 * Country: <multi line of data>
	 * Custom1: <multi line of data>
	 * Custom2: <multi line of data>
	 * Custom3: <multi line of data>
	 * Custom4: <multi line of data>
	 * Note: <multi-line data>
	 * 
	 * Name strings are case sensitive.  All name strings are preceded with
	 * PREFIX (defined above) to identify them as true tokens.
	 */
	DBGPrintf(DBG_INFO,("Starting to process data file.\n"));
	while( cPtr < (int)((int)file_text + filelen) )
	{
		/* Move past the record prefix */
		cPtr = cPtr + strlen(PREFIX);

		/*
		 * Get the current name, value pair.
		 */
		name=strtok((char *)cPtr, ":");

		cPtr = cPtr + strlen(name) + 2;

		/* Grab the multiple lines of a "value". */
		value = (char *)cPtr;
		while ( 
			(strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 ) && 
			((int)cPtr < (int)((int)file_text + filelen) )
			)
				cPtr++;

		/* Remove trailing newline */
		endPtr = cPtr - 1;
				
		/*
		 * Null terminate the string 
		 */
		strncpy((char *)endPtr, "\0", 1);

		if ( strncmp(value, "(null)", 6) == 0 )
			value = "";

		DBGPrintf(DBG_INFO,("value: *%s*\n", value));
		switch (count)
		{
			case 0: 
				current_link->category = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->category, value);
				break;
			case 1: 
				current_link->recordid = atoi(value);
				break;
			case 2: 
				current_link->listPhone = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->listPhone, value);
				break;
			case 3: 
				current_link->attributes = atoi(value);
				break;

			/* Name, title, company fields */
			case 4: 
				current_link->last = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->last, value);
				break;
			case 5: 
				current_link->first = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->first, value);
				break;
			case 6: 
				current_link->title = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->title, value);
				break;
			case 7: 
				current_link->company = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->company, value);
				break;

			/* Phone / email / Fax / Other fields */
			case 8: 
				current_link->phoneLabel1 = (char *)XPSafeAlloc(strlen(name) + 1);
				strcpy(current_link->phoneLabel1, name);
				current_link->phone1 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->phone1, value);
				break;
			case 9: 
				current_link->phoneLabel2 = (char *)XPSafeAlloc(strlen(name) + 1);
				strcpy(current_link->phoneLabel2, name);
				current_link->phone2 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->phone2, value);
				break;
			case 10: 
				current_link->phoneLabel3 = (char *)XPSafeAlloc(strlen(name) + 1);
				strcpy(current_link->phoneLabel3, name);
				current_link->phone3 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->phone3, value);
				break;
			case 11: 
				current_link->phoneLabel4 = (char *)XPSafeAlloc(strlen(name) + 1);
				strcpy(current_link->phoneLabel4, name);
				current_link->phone4 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->phone4, value);
				break;
			case 12: 
				current_link->phoneLabel5 = (char *)XPSafeAlloc(strlen(name) + 1);
				strcpy(current_link->phoneLabel5, name);
				current_link->phone5 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->phone5, value);
				break;

			/* Addr, City, State, Zip, Country fields */
			case 13: 
				current_link->address = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->address, value);
				break;
			case 14: 
				current_link->city = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->city, value);
				break;
			case 15: 
				current_link->state = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->state, value);
				break;
			case 16: 
				current_link->zip = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->zip, value);
				break;
			case 17: 
				current_link->country = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->country, value);
				break;

			/* Custom fields */
			case 18: 
				current_link->custom1 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->custom1, value);
				break;
			case 19: 
				current_link->custom2 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->custom2, value);
				break;
			case 20: 
				current_link->custom3 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->custom3, value);
				break;
			case 21: 
				current_link->custom4 = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->custom4, value);
				break;

			/* Note field */
			case 22: 
				current_link->note = (char *)XPSafeAlloc(strlen(value) + 1);
				strcpy(current_link->note, value);
				break;

		}

		count++;

		/*
		 * If we've processed all the entries for one record and
		 * we're not at the end of the data, allocate a new link.
		 */
		if ( ( count == 23 ) && ( cPtr < (int)((int)file_text + filelen) ) )
		{

			DBGPrintf(DBG_INFO,("Record processed, allocating new link.\n"));

			/* Allocate a new link for list */
			pentry = (PilotAddr_T *)XPSafeAlloc(sizeof(PilotAddr_T));
			memset(pentry,0,sizeof(PilotAddr_T));
		
			current_link->next = pentry;
			current_link = pentry;
			current_link->next = NULL;

			/* Reset the counter */
			count = 0;

		}

	}

	/*
	 * Sort the link list.
	 */
	SortAddresses();

	/*
	 * Generate the link list of categories.
	 */
	SortCategoryList(&Pilot_Addr_Categories);

#ifdef DEBUG
	PrintFields();
#endif
	return ( XP_E_SUCCESS );
}

static void
PrintFields()
{
#ifdef DEBUG
	char		*fname="PrintFields()";
#endif

	PilotAddr_T		*current_link;
	
	DBGEnter();

	current_link = Pilot_Address;
	while (current_link != NULL )
	{
		DBGPrintf(DBG_TRACE3, ("New record: \n"));
		DBGWrite(DBG_TRACE3,
			("\tlast, first: %s, %s\n", current_link->last, current_link->first));
		DBGWrite(DBG_TRACE3,
			("\tcategory   : %s\n", current_link->category));
		DBGWrite(DBG_TRACE3,
			("\ttitle      : %s\n", current_link->title));
		DBGWrite(DBG_TRACE3,
			("\tcompany    : %s\n", current_link->company));
		DBGWrite(DBG_TRACE3,
			("\t%s       : %s\n",current_link->phoneLabel1, current_link->phone1));
		DBGWrite(DBG_TRACE3,
			("\t%s       : %s\n",current_link->phoneLabel2, current_link->phone2));
		DBGWrite(DBG_TRACE3,
			("\t%s       : %s\n",current_link->phoneLabel3, current_link->phone3));
		DBGWrite(DBG_TRACE3,
			("\t%s       : %s\n",current_link->phoneLabel4, current_link->phone4));
		DBGWrite(DBG_TRACE3,
			("\t%s       : %s\n",current_link->phoneLabel5, current_link->phone5));
		DBGWrite(DBG_TRACE3,
			("\taddress    : %s\n", current_link->address));
		DBGWrite(DBG_TRACE3,
			("\tcity       : %s\n", current_link->city));
		DBGWrite(DBG_TRACE3,
			("\tstate      : %s\n", current_link->state));
		DBGWrite(DBG_TRACE3,
			("\tzip        : %s\n", current_link->zip));
		DBGWrite(DBG_TRACE3,
			("\tcountry    : %s\n", current_link->country));
		DBGWrite(DBG_TRACE3,
			("\tcustom1    : %s\n", current_link->custom1));
		DBGWrite(DBG_TRACE3,
			("\tcustom2    : %s\n", current_link->custom2));
		DBGWrite(DBG_TRACE3,
			("\tcustom3    : %s\n", current_link->custom3));
		DBGWrite(DBG_TRACE3,
			("\tcustom4    : %s\n", current_link->custom4));
		DBGWrite(DBG_TRACE3,
			("\tnote       : %s\n", current_link->note));
/*
		DBGWrite(DBG_TRACE3,
			("\t     length: %d\n", strlen(current_link->note)));
		DBGWrite(DBG_TRACE3,
			("\tAttributes : %d\n", current_link->attributes));
			pentry->category = NULL;
*/

		current_link = current_link->next;
	}
}



/*========================================================================
 *	Name:			SortAddresses
 *	Prototype:	void SortAddresses()
 *					
 *	Description:
 * Sort the address list based on which field is of interest.
 *
 *	Input Arguments:
 *
 * Notes:
 * Uses a simple bubble sort, which should be sufficient for most address
 * lists.  We first sort on the last name, then run the list again to sort
 * by first name.  This only really works if the data in the Pilot is
 * correctly entered (ie the Last Name entry in the pilot contains *only*
 * the Last Name and *not* the last name and first name seperated by a
 * comma).
 *========================================================================*/
static void
SortAddresses()
{
#ifdef DEBUG
	char				*fname="SortAddresses()";
#endif

	PilotAddr_T		*current_link, *prev_link;

	gboolean			repeat, rc1, rc2;

	DBGEnter();

repeat_primary:
	repeat = False;
	current_link = Pilot_Address;
	prev_link = NULL;
	while (current_link != NULL)
	{
		rc1 = False;
		if ( current_link->next != NULL )
			rc1 = SortFields(SORT_BY_LAST, current_link, prev_link);
		if ( rc1 )
			repeat = True;
		prev_link = current_link;
		current_link = current_link->next;
	}
	if ( repeat )
		goto repeat_primary;

repeat_secondary:
	repeat = False;
	current_link = Pilot_Address;
	prev_link = NULL;
	while (current_link != NULL)
	{
		rc2 = False;
		if ( current_link->next != NULL )
			if ( (current_link->last != NULL)&&(current_link->next->last != NULL) )
				if ( strcmp(current_link->last, current_link->next->last) == 0 )
					rc2 = SortFields(SORT_BY_FIRST, current_link, prev_link);
		if ( rc2 )
			repeat = True;
		prev_link = current_link;
		current_link = current_link->next;
	}
	if ( repeat )
		goto repeat_secondary;
}



/*========================================================================
 *	Name:			SortFields
 *	Prototype:	static Boolean SortFields(int field)
 *					
 *	Description:
 * Sort two entries.
 *
 *	Input Arguments:
 *		int				field					field to sort on
 *		PilotAddr_T		*current_link		current link to sort
 *
 * Returns
 *		True				switch occured
 *		False				switch did not occur
 *
 * Notes:
 *	Assumes current_link and current_link->next exist (are non-NULL)
 *========================================================================*/
static gboolean
SortFields(
	int				field,
	PilotAddr_T		*current_link,
	PilotAddr_T		*prev_link
)
{
#ifdef DEBUG
	char				*fname="SortFields()";
#endif

	PilotAddr_T		*next_link;

	gboolean			swap;

	DBGEnter();

	swap = False;
	next_link = current_link->next;
	if ( next_link == NULL )
		return(False);

	switch ( field )
	{
		case SORT_BY_LAST:
			if ( (current_link->last != NULL) && (next_link->last == NULL) )
					swap = True;
			if ( (current_link->last != NULL) && (next_link->last != NULL) )
				if ( strcmp(current_link->last, next_link->last) > 0 )
					swap = True;
			break;
		case SORT_BY_FIRST:
			if ( (current_link->first != NULL) && (next_link->first == NULL) )
					swap = True;
			if ( (current_link->first != NULL) && (next_link->first != NULL) )
				if ( strcmp(current_link->first, next_link->first) > 0 )
					swap = True;
			break;
		case SORT_BY_COMPANY:
			if ( (current_link->company != NULL) && (next_link->company == NULL) )
					swap = True;
			if ( (current_link->company != NULL) && (next_link->company != NULL) )
				if ( strcmp(current_link->company, next_link->company) > 0 )
					swap = True;
			break;
		case SORT_BY_CATEGORY:
			if ( (current_link->category != NULL) && (next_link->category == NULL) )
					swap = True;
			if ( (current_link->category != NULL) && (next_link->category != NULL) )
				if ( strcmp(current_link->category, next_link->category) > 0 )
					swap = True;
			break;
		case SORT_BY_CITY:
			if ( (current_link->city != NULL) && (next_link->city == NULL) )
					swap = True;
			if ( (current_link->city != NULL) && (next_link->city != NULL) )
				if ( strcmp(current_link->city, next_link->city) > 0 )
					swap = True;
			break;
	}
	if ( swap )
	{
		if ( prev_link != NULL )
			prev_link->next = next_link;
		else
			Pilot_Address = next_link;
		current_link->next = next_link->next;
		next_link->next = current_link;
		return(True);
	}
	return(False);
}


/*========================================================================
 *	Name:			XPFreeAddressList
 *	Prototype:	void XPFreeAddressList(PilotAddr_T	**head)
 *					
 *	Description:
 * Free up the link list of addresses.
 *
 * Notes:
 *========================================================================*/
void
XPFreeAddressList(
	PilotAddr_T			**head
)
{
	PilotAddr_T			*current_link, *next_link;

	/*
	 * Free up any old links in the list.
	 */
	current_link = *head;
	while (current_link != NULL )
	{
		/* Save the next link */
		next_link = current_link->next;

		/* Free up any allocated storage within the link */
		if ( current_link->last != (char *)NULL )
			free(current_link->last);

		if ( current_link->first != (char *)NULL )
			free(current_link->first);

		if ( current_link->title != (char *)NULL )
			free(current_link->title);

		if ( current_link->company != (char *)NULL )
			free(current_link->company);

		if ( current_link->phoneLabel1 != (char *)NULL )
			free(current_link->phoneLabel1);

		if ( current_link->phone1 != (char *)NULL )
			free(current_link->phone1);

		if ( current_link->phoneLabel2 != (char *)NULL )
			free(current_link->phoneLabel2);

		if ( current_link->phone2 != (char *)NULL )
			free(current_link->phone2);

		if ( current_link->phoneLabel3 != (char *)NULL )
			free(current_link->phoneLabel3);

		if ( current_link->phone3 != (char *)NULL )
			free(current_link->phone3);

		if ( current_link->phoneLabel4 != (char *)NULL )
			free(current_link->phoneLabel4);

		if ( current_link->phone4 != (char *)NULL )
			free(current_link->phone4);

		if ( current_link->phoneLabel5 != (char *)NULL )
			free(current_link->phoneLabel5);

		if ( current_link->phone5 != (char *)NULL )
			free(current_link->phone5);

		if ( current_link->address != (char *)NULL )
			free(current_link->address);

		if ( current_link->city != (char *)NULL )
			free(current_link->city);

		if ( current_link->state != (char *)NULL )
			free(current_link->state);

		if ( current_link->zip != (char *)NULL )
			free(current_link->zip);

		if ( current_link->country != (char *)NULL )
			free(current_link->country);

		if ( current_link->custom1 != (char *)NULL )
			free(current_link->custom1);

		if ( current_link->custom2 != (char *)NULL )
			free(current_link->custom2);

		if ( current_link->custom3 != (char *)NULL )
			free(current_link->custom3);

		if ( current_link->custom4 != (char *)NULL )
			free(current_link->custom4);

		if ( current_link->note != (char *)NULL )
			free(current_link->note);

		if ( current_link->category != (char *)NULL )
			free(current_link->category);


		/* free up the current link */
		free(current_link);
		current_link = next_link;
	}
	head=NULL;
}


/*========================================================================
 *	Name:			XPFreeCategoryList
 *	Prototype:	static void XPFreeCategoryList(PilotCategory_T **head)
 *					
 *	Description:
 * Free up the link list of categories.
 *
 * Notes:
 *========================================================================*/
void
XPFreeCategoryList(
	PilotCategory_T	**head
)
{
	PilotCategory_T		*category_link, *next_link;

	category_link = *head;
	while (category_link != NULL )
	{
		if ( category_link->name != NULL )
			free(category_link->name);
		next_link = category_link->next;
		free(category_link);
		category_link = next_link;
	}

	/*
	 * Reset the head link for the appropriate Category list.
	 */
	*head = NULL;

}


/*========================================================================
 *	Name:			SortCategoryList
 *	Prototype:	static void SortCategoryList(PilotCategory_T **head)
 *					
 *	Description:
 * Sort the list of categories.
 *
 * Notes:
 *========================================================================*/
static void
SortCategoryList(
	PilotCategory_T	**head
)
{
	gboolean			repeat;
	PilotCategory_T		*category_link, *prev_link, *next_link;

repeat_category:
	repeat = False;
	category_link = *head;
	prev_link = NULL;
	while (category_link != NULL)
	{
		next_link = category_link->next;
		if ( next_link != NULL )
		{
			if ( strcmp(category_link->name, next_link->name) > 0 )
			{
				if ( prev_link != NULL )
					prev_link->next = next_link;
				else
					*head = next_link;
				category_link->next = next_link->next;
				next_link->next = category_link;
				repeat = True;
			}
		}
		prev_link = category_link;
		category_link = category_link->next;
	}
	if ( repeat )
		goto repeat_category;

}


/*========================================================================
 *	Name:			GenerateCategoryList
 *	Prototype:	static void GenerateCategoryList()
 *					
 *	Description:
 * Generate the link list of categories for the given link list.
 *
 * Notes:
 *========================================================================*/
static void
GenerateCategoryList(
	char					*category_name,
	PilotCategory_T	**head
)
{
#ifdef DEBUG
	char							*fname="GenerateCategoryList()";
#endif

	PilotCategory_T			*category_link, *last_cat_link=NULL;

	gboolean						found_match;

	DBGEnter();

	if ( *head == NULL )
	{
		/* First entry:  allocate an entry and point the header to it. */
		*head = 
			(PilotCategory_T *)XPSafeAlloc(sizeof(PilotCategory_T));
		memset(*head,0,sizeof(PilotCategory_T));
		(*head)->next = NULL;
		(*head)->name = XPSafeAlloc(strlen(category_name)+1);
		strcpy((*head)->name, category_name);
	}
	else
	{
		/* Check to see if category already exists in list */
		category_link = *head;
		found_match = False;
		while (category_link != NULL )
		{
			if ( strcmp(category_link->name, category_name) == 0 )
			{
				found_match = True;
				break;
			}
			last_cat_link = category_link;
			category_link = category_link->next;
		}

		/*
		 * If no match was found, insert a new link at the end of the
		 * current list.  We'll sort this list later.
		 */
		if (!found_match)
		{
			/*
			 * last_cat_link should point to the last link in the current
			 * list.
			 */
			last_cat_link->next = 
				(PilotCategory_T *)XPSafeAlloc(sizeof(PilotCategory_T));
			memset(last_cat_link->next,0,sizeof(PilotCategory_T));
			category_link = last_cat_link->next;
			category_link->next = NULL;
			category_link->name = XPSafeAlloc(strlen(category_name)+1);
			strcpy(category_link->name, category_name);
		}
	}
}


/*========================================================================
 *	Name:			XPPilotParseTodo
 *	Prototype:	void XPPilotParseTodo(char *filename)
 *					
 *	Description:
 *	Parses the Todo List flat file into a link list.  Also builds the
 * category link list for the Todo List.
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *		XP_E_FILE_EXIST		file to parse doesn't exist or is not a text file
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *	This version uses the new pseudo-XML format.  
 *		
 *========================================================================*/
int 
XPPilotParseTodo(
	char	*filename
)
{
#ifdef DEBUG
	char				*fname="XPPiloParseTodo()";
#endif

	PilotTodo_T				*pentry;
	static PilotTodo_T	*current_link=NULL;

	struct stat		stat_buf;
	FILE				*fd;
	char				*file_text;
	int				filelen;
	int				cPtr, endPtr;
	int				count;
	char				*name, *value;


	DBGEnter();
	DBGPrintf(DBG_INFO,("pilot db to look for: %s\n", filename));

	/*
	 * Verify file exists.  Return if it doesn't.
	 */
	if ( stat(filename, &stat_buf) != 0 )
		return XP_E_FILE_EXIST;

	if ( ! ( stat_buf.st_mode & S_IFREG ) )
		return XP_E_FILE_EXIST;

	/*
	 * File is ok.  Open it and read in the file.
	 */
	fd = fopen(filename, "r");

	/*
	 * Free up any old link lists.
	 */
	FreeTodoList();
	XPFreeCategoryList(&Pilot_Todo_Categories);

	/* Read the file into memory - it's faster this way. */
	fseek(fd, 0, SEEK_END);
	filelen = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	file_text = (char*)malloc(filelen+1);
	if (file_text == NULL) 
	{
		/* This should never happen. */
		perror("malloc()");
		exit(1);
	}
   fread(file_text, filelen, 1, fd);


	/*
	 * Initialize pointers, counters used to parse input text.
	 */
	cPtr = (int)file_text;
	count = 0;

	/*
	 * Move past the Magic Mark
	 */
	name=strtok((char *)cPtr, "\n");
	cPtr = cPtr + strlen(name) + 1;

	/*
	 * Skip the category fields from the file first.
	 * When we're done, move past the marker that separates the category
	 * list from the records.
	 */
	while ( 
		(cPtr < (int)((int)file_text + filelen) ) &&
		( strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 )
	)
	{
		name=strtok((char *)cPtr, "\n");
		GenerateCategoryList(name, &Pilot_Todo_Categories);
		cPtr = cPtr + strlen(name) + 1;
	}

	/*
	 * Allocate a new link for an address record if there are any records.
	 */
	if (cPtr < (int)((int)file_text + filelen) )
	{
		pentry = (PilotTodo_T *)XPSafeAlloc(sizeof(PilotTodo_T));
		memset(pentry,0,sizeof(PilotTodo_T));

		Pilot_Todo = pentry;
		current_link = pentry;
		current_link->next = NULL;
	}

	/* 
	 * Parse the text from the input file.  The format of each entry in
	 * the input files must be in the following name/value formats:
	 *
	 * Category: <single line of data>
	 * RecordID: <single line of data>
	 * Priority: <single line of data>
	 * Completed: <single line of data>
	 * Due: <single line of data>
	 * Description: <multi-line data>
	 * Note: <multi-line data>
	 * <blank line>
	 * 
	 * Name strings are case insensitive.
	 */

	while( cPtr < (int)((int)file_text + filelen) )
	{

		/* Move past the record prefix */
		cPtr = cPtr + strlen(PREFIX);

		/*
		 * Get the current name, value pair.
		 */
		name=strtok((char *)cPtr, ":");
		cPtr = cPtr + strlen(name) + 2;
	
		/* Grab the multiple lines of a "value". */
		value = (char *)cPtr;
		while ( 
			(strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 ) && 
			((int)cPtr < (int)((int)file_text + filelen) )
			)
				cPtr++;
	
		/* Remove trailing newline */
		endPtr = cPtr - 1;
				
		/*
		 * Null terminate the string 
		 */
		strncpy((char *)endPtr, "\0", 1);


		/* Note field */
		if ( strncasecmp(name, "Note", 4) == 0 )
		{
			current_link->note = XPSafeAlloc(strlen(value) + 1);
			strcpy(current_link->note, value);
		}

		/* Category field */
		if ( strncasecmp(name, "Category", 8) == 0 )
		{
			current_link->category = XPSafeAlloc(strlen(value) + 1);
			strcpy(current_link->category, value);
		}

		/* Skip RecordID field */
		/*
		if ( strncasecmp(name, "RecordID", 8) == 0 )
		{
			continue;
		}
		*/

		/* Skip Attributes field */
		/*
		if ( strncasecmp(name, "Attribute", 9) == 0 )
		{
			continue;
		}
		*/

		/* Description field */
		if ( strncasecmp(name, "Description", 11) == 0 )
		{
			current_link->title = XPSafeAlloc(strlen(value) + 1);
			strcpy(current_link->title, value);
		}

		/* Priority field */
		if ( strncasecmp(name, "Priority", 8) == 0 )
		{
			current_link->priority = atoi(value);
		}

		/* For "Complete", if it's "Yes", it's 1, otherwise it's 0. */
		if ( strncasecmp(name, "Completed", 9) == 0 )
		{
			if ( strncasecmp(value, "Yes", 3) == 0 )
			{
				current_link->completed = 1;
			}
			else
			{
				current_link->completed = 0;
			}
		}

		/* Due Date field */
		if ( strncasecmp(name, "Due", 3) == 0 )
		{
			if ( strncasecmp(value, "no date", 7) == 0 )
			{
				current_link->indefinite = 1;
			}
			else
			{
				current_link->indefinite = 0;
			}
			current_link->due_date = XPSafeAlloc(strlen(value) + 1);
			strcpy(current_link->due_date, value);
		}

		/* Increment the line of the current entry we are on. */
		count++;

		if ( ( count == 8 ) && ( cPtr < (int)((int)file_text + filelen) ) )
		{
			DBGPrintf(DBG_INFO,("Allocating new link for Todos\n"));
			/*
			 * Allocate a new link for an Todo record.
			 */
			pentry = (PilotTodo_T *)XPSafeAlloc(sizeof(PilotTodo_T));
			memset(pentry,0,sizeof(PilotTodo_T));
		
			current_link->next = pentry;
			current_link = pentry;
			current_link->next = NULL;


			/* Reset the line counter */
			count = 0;
		}
	}

	/*
	 * Now build the category list.
	 */
	SortCategoryList(&Pilot_Todo_Categories);

	/* Cleanup */
	free(file_text);
	return ( XP_E_SUCCESS );
}


/*========================================================================
 *	Name:			FreeTodoList
 *	Prototype:	static void FreeTodoList()
 *					
 *	Description:
 * Free up the link list of Todo entries.
 *
 * Notes:
 *========================================================================*/
static void
FreeTodoList()
{

	PilotTodo_T			*current_link, *next_link;

	/*
	 * Free up any old links in the list.
	 */
	current_link = Pilot_Todo;
	while (current_link != NULL )
	{
		/* Save the next link */
		next_link = current_link->next;

		/* Free up any allocated storage within the link */
		if ( current_link->category != NULL )
			free(current_link->category);
		if ( current_link->due_date != NULL )
			free(current_link->due_date);
		if ( current_link->title != NULL )
			free(current_link->title);
		if ( current_link->note != NULL )
			free(current_link->note);

		/* free up the current link */
		free(current_link);
		current_link = next_link;
	}
	Pilot_Todo=NULL;
}


/*========================================================================
 *	Name:			XPWriteTodos()
 *	Prototype:	void XPWriteTodos()
 *					
 *	Description:
 *	Write the todo list back to disk.
 *
 *	Input Arguments:
 *		char	*filename		name of file to write to
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *		
 *========================================================================*/
void 
XPWriteTodos(
	char	*filename
)
{
	FILE					*fp;
	char					tmpfile[MAXPATHLEN];
	mode_t				oldmask;
	PilotTodo_T			*current_link;
	PilotCategory_T	*category_link;
	char					cmd[MAXPATHLEN];
	char					magicmark[0xffff];

	/*
	 * Open a temporary file - make sure it's only readable by the user.
	 */
   sprintf(tmpfile,"%s/xp3.%d", app_res.note_dir, (int)getpid());
	oldmask = umask((mode_t)022);

   /*
    * Open the first temporary file for writing
    */
	if ((fp = fopen(tmpfile, "w")) == NULL) 
	{
		/* Not likely, but you never know */
		fprintf(stderr, "xnotes: ");
		perror(tmpfile);
		return;
	}

	/* Write out the magic string */
	sprintf(magicmark, MAGICMARK, "todos");
	fprintf(fp, "%s\n", magicmark);

	/*
	 * Write out the categories first. 
	 */
	category_link = Pilot_Todo_Categories;
	while (category_link != NULL)
	{
		fprintf(fp, "%s\n", category_link->name);
		category_link = category_link->next;
	}


	/*
	 * Start with the first link.
	 */
	current_link = Pilot_Todo;

	/*
	 * For each link, write out the entry.
	 */
	while (current_link != NULL)
	{

		/* Skip these for now - we'll manage them in a later version */
		fprintf(fp, "%sAttribute: 0\n", PREFIX);
		fprintf(fp, "%sRecordID: 0\n", PREFIX);

		fprintf(fp, "%sCategory: %s\n", PREFIX, current_link->category);
		fprintf(fp, "%sPriority: %d\n", PREFIX, current_link->priority);
		fprintf(fp, "%sCompleted: %s\n", PREFIX, 
			current_link->completed ? "Yes" : "No");

		if(current_link->indefinite)
			fprintf(fp, "%sDue: No Date\n", PREFIX);
		else
			fprintf(fp, "%sDue: %s\n", PREFIX, current_link->due_date);

		if(current_link->title)
			fprintf(fp, "%sDescription: %s\n", PREFIX, current_link->title);
		else
			fprintf(fp, "%sDescription: \n", PREFIX);

		if(current_link->note)
			fprintf(fp, "%sNote: %s\n", PREFIX, current_link->note );
		else
			fprintf(fp, "%sNote: \n", PREFIX);

		current_link = current_link->next;
	}

	/*
	 * Close the file and copy it over the original.
	 * This is to help prevent partial overwrite during system crashes.
	 * It's not foolproof, but what file operation is atomic?
	 */
	fclose(fp);
	if ( rename(tmpfile, filename) != 0 )
	{
		sprintf(cmd, "xnotes: %s", tmpfile);
		perror(cmd);
		sprintf(cmd,"Couldn't rename %s\nto %s.\n"
			"Save of Todo List failed.",
			tmpfile, filename);
		XPErrPopUp(cmd);
		umask((mode_t)oldmask);
		return;
	}

	/* Reset the file permissions mask */
	umask((mode_t)oldmask);

}


/*========================================================================
 *	Name:			XPPilotParseMemos
 *	Prototype:	void XPPilotParseMemos(char *filename)
 *					
 *	Description:
 *	Parses the Memos List flat file into a link list.  Also builds the
 * category link list for the Memos List.
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *		XP_E_FILE_EXIST		file to parse doesn't exist or is not a text file
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *	This version uses the new pseudo-XML format.  
 *		
 *========================================================================*/
int 
XPPilotParseMemos(
	char	*filename
)
{
#ifdef DEBUG
	char				*fname="XPPiloParseMemos()";
#endif

	PilotMemo_T				*pentry;
	static PilotMemo_T	*current_link=NULL;

	struct stat		stat_buf;
	FILE				*fd;
	char				*file_text;
	int				filelen;
	int				cPtr, endPtr;
	int				count;
	char				*name, *value;
	int				xnoteid;


	DBGEnter();
	DBGPrintf(DBG_INFO,("pilot db to look for: %s\n", filename));

	/*
	 * Verify file exists.  Return if it doesn't.
	 */
	if ( stat(filename, &stat_buf) != 0 )
		return XP_E_FILE_EXIST;

	if ( ! ( stat_buf.st_mode & S_IFREG ) )
		return XP_E_FILE_EXIST;

	/*
	 * File is ok.  Open it and read in the file.
	 */
	fd = fopen(filename, "r");

	/*
	 * Free up any old link lists.
	 */
	XPFreeMemoList(&Pilot_Memo);
	XPFreeCategoryList(&Pilot_Memo_Categories);

	/*
	 * Allocate a new link for an address record.
	 */
	pentry = (PilotMemo_T *)XPSafeAlloc(sizeof(PilotMemo_T));
	memset(pentry,0,sizeof(PilotMemo_T));

	Pilot_Memo = pentry;
	current_link = pentry;
	current_link->next = NULL;

	/* Read the file into memory - it's faster this way. */
	fseek(fd, 0, SEEK_END);
	filelen = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	file_text = (char*)malloc(filelen+1);
	if (file_text == NULL) 
	{
		/* This should never happen. */
		perror("malloc()");
		exit(1);
	}
   fread(file_text, filelen, 1, fd);
	fclose(fd);


	/*
	 * Initialize pointers, counters used to parse input text.
	 */
	cPtr = (int)file_text;
	count = 0;
	xnoteid = 0; /* Keeps a local id for each record */

	/*
	 * Move past the Magic Mark
	 */
	name=strtok((char *)cPtr, "\n");
	cPtr = cPtr + strlen(name) + 1;

	/*
	 * Skip the category fields from the file first.
	 * When we're done, move past the marker that separates the category
	 * list from the records.
	 */
	while( strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 )
	{
		name=strtok((char *)cPtr, "\n");
		cPtr = cPtr + strlen(name) + 1;
	}

	/* 
	 * Parse the text from the input file.  The format of each entry in
	 * the input files must be in the following name/value formats:
	 *
	 * Category: <single line of data>
	 * RecordID: <single line of data>
	 * Priority: <single line of data>
	 * Completed: <single line of data>
	 * Due: <single line of data>
	 * Description: <multi-line data>
	 * Note: <multi-line data>
	 * <blank line>
	 * 
	 * Name strings are case insensitive.
	 */

	while( cPtr < (int)((int)file_text + filelen) )
	{

		/* Move past the record prefix */
		cPtr = cPtr + strlen(PREFIX);

		/*
		 * Get the current name, value pair.
		 */
		name=strtok((char *)cPtr, ":");
		cPtr = cPtr + strlen(name) + 2;
	
		/* Grab the multiple lines of a "value". */
		value = (char *)cPtr;
		while ( 
			(strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 ) && 
			((int)cPtr < (int)((int)file_text + filelen) )
			)
				cPtr++;
	
		/* Remove trailing newline */
		endPtr = cPtr - 1;
				
		/*
		 * Null terminate the string 
		 */
		strncpy((char *)endPtr, "\0", 1);


		/* Skip RecordID field */
		/*
		if ( strncasecmp(name, "RecordID", 8) == 0 )
		{
			continue;
		}
		*/

		/* Skip Attributes field */
		/*
		if ( strncasecmp(name, "Attribute", 9) == 0 )
		{
			continue;
		}
		*/

		/* Category field */
		if ( strncasecmp(name, "Category", 8) == 0 )
		{
			current_link->category = XPSafeAlloc(strlen(value) + 1);
			strcpy(current_link->category, value);
		}

		/* Title field */
		if ( strncasecmp(name, "Title", 11) == 0 )
		{
			current_link->title = XPSafeAlloc(strlen(value) + 1);
			strcpy(current_link->title, value);
		}

		/* Text field */
		if ( strncasecmp(name, "Text", 4) == 0 )
		{
			current_link->text = XPSafeAlloc(strlen(value) + 1);
			strcpy(current_link->text, value);
		}

		/* Increment the line of the current entry we are on. */
		count++;

		if ( count == 5 ) 
		{
			/* Save the local id for this record */
			current_link->xnotesid = xnoteid;
			xnoteid++;

			if ( ( cPtr < (int)((int)file_text + filelen) ) )
			{
				/*
				 * Allocate a new link for a memo record.
				 */
				pentry = (PilotMemo_T *)XPSafeAlloc(sizeof(PilotMemo_T));
				memset(pentry,0,sizeof(PilotMemo_T));
			
				current_link->next = pentry;
				current_link = pentry;
				current_link->next = NULL;

			}

			/* Reset the line counter */
			count = 0;
		}
	}

	/*
	 * Now build the category list.
	 */
	XPGenerateMemoCategoryList(Pilot_Memo, &Pilot_Memo_Categories);
	SortCategoryList(&Pilot_Memo_Categories);

	/* Cleanup */
	free(file_text);
	return ( XP_E_SUCCESS );
}


/*========================================================================
 *	Name:			XPFreeMemoList
 *	Prototype:	void XPFreeMemoList((PilotMemo_T **)head)
 *					
 *	Description:
 * Free up a link list of Memo entries.
 *
 * Notes:
 * This one is known externally so we can free up a temporary list created
 * when preparing to upload XNotes to the Pilot Memo's db.
 *========================================================================*/
void
XPFreeMemoList(
	PilotMemo_T		**head
)
{

	PilotMemo_T			*current_link, *next_link;

	/*
	 * Free up any old links in the list.
	 */
	current_link = *head;
	while (current_link != NULL )
	{
		/* Save the next link */
		next_link = current_link->next;

		/* Free up any allocated storage within the link */
		if ( current_link->category != NULL )
			free(current_link->category);
		if ( current_link->title != NULL )
			free(current_link->title);
		if ( current_link->text != NULL )
			free(current_link->text);

		/* free up the current link */
		free(current_link);
		current_link = next_link;
	}
	head=NULL;
}


/*========================================================================
 *	Name:			XPGenerateMemoCategoryList
 *	Prototype:	void XPGenerateMemoCategoryList()
 *					
 *	Description:
 * Generate the link list of memo categories.  This one is slightly
 * different than the Address and Todo function because it gets called from
 * external routines and is used slightly differently.
 *
 * Notes:
 *========================================================================*/
void
XPGenerateMemoCategoryList(
	PilotMemo_T					*mhead,
	PilotCategory_T			**chead
)
{
#ifdef DEBUG
	char							*fname="XPGenerateMemoCategoryList()";
#endif

	PilotMemo_T					*memo_link;
	PilotCategory_T			*category_link, *last_cat_link=NULL;

	gboolean						found_match;

	DBGEnter();

	memo_link = mhead;
	while (memo_link != NULL )
	{
		if ( *chead == NULL )
		{
			/* First entry:  allocate an entry and point the header to it. */
			*chead = 
				(PilotCategory_T *)XPSafeAlloc(sizeof(PilotCategory_T));
			(PilotCategory_T *)(*chead)->next = NULL;
			(PilotCategory_T *)(*chead)->name = 
				XPSafeAlloc(strlen(memo_link->category)+1);
			strcpy((*chead)->name, memo_link->category);
		}
		else
		{
			/* Check to see if category already exists in list */
			category_link = (PilotCategory_T *)(*chead);
			found_match = False;
			while (category_link != NULL )
			{
				if ( strcmp(category_link->name, memo_link->category) == 0 )
				{
					found_match = True;
					break;
				}
				last_cat_link = category_link;
				category_link = category_link->next;
			}

			/*
			 * If no match was found, insert a new link at the end of the
			 * current list.  We'll sort this list later.
			 */
			if (!found_match)
			{
				/*
				 * last_cat_link should point to the last link in the current
				 * list.
				 */
				last_cat_link->next = 
					(PilotCategory_T *)XPSafeAlloc(sizeof(PilotCategory_T));
				category_link = last_cat_link->next;
				category_link->next = NULL;
				category_link->name = XPSafeAlloc(strlen(memo_link->category)+1);
				strcpy(category_link->name, memo_link->category);
			}
		}

		/*
		 * Get the next link in the Memo list.
		 */
		memo_link = memo_link->next;
	}
}


/*========================================================================
 *	Name:			XPWriteMemos()
 *	Prototype:	void XPWriteMemos(PilotMemo_T *mhead, PilotCategory_T *chead, char *filename)
 *					
 *	Description:
 *	Write a memo link list to disk.
 *
 *	Input Arguments:
 *		PilotMemo_T				*mhead		head of link list to process
 *		PilotCategory_T		*chead		head of link list to process
 *		char						*filename	name of temporary file to write to
 *
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *		
 *========================================================================*/
void 
XPWriteMemos(
	PilotMemo_T			*mhead,
	PilotCategory_T	*chead,
	char					*filename
)
{
	FILE					*fp;
	mode_t				oldmask;
	PilotMemo_T			*current_link;
	PilotCategory_T	*category_link;
	char					magicmark[0xffff];

	/*
	 * Open a temporary file - make sure it's only readable by the user.
	 */
	oldmask = umask((mode_t)022);

   /*
    * Open the first temporary file for writing
    */
	if ((fp = fopen(filename, "w")) == NULL) 
	{
		/* Not likely, but you never know */
		fprintf(stderr, "xnotes: ");
		perror(filename);
		return;
	}

	/* Write out the magic string */
	sprintf(magicmark, MAGICMARK, "todos");
	fprintf(fp, "%s\n", magicmark);

	/*
	 * Write out the categories first. 
	 */
	category_link = chead;
	while (category_link != NULL)
	{
		fprintf(fp, "%s\n", category_link->name);
		category_link = category_link->next;
	}


	/*
	 * Start with the first link.
	 */
	current_link = mhead;

	/*
	 * For each link, write out the entry.
	 */
	while (current_link != NULL)
	{

		/* Skip these for now - we'll manage them in a later version */
		fprintf(fp, "%sRecordID: 0\n", PREFIX);
		fprintf(fp, "%sAttributes: 0\n", PREFIX);
		fprintf(fp, "%sCategory: %s\n", PREFIX, current_link->category);
		fprintf(fp, "%sTitle: %s\n", PREFIX, current_link->title);
		if(current_link->text)
			fprintf(fp, "%sText: %s\n", PREFIX, current_link->text );
		else
			fprintf(fp, "%sText: \n", PREFIX);

		current_link = current_link->next;
	}

	/*
	 * Close the file.
	 */
	fclose(fp);

	/* Reset the file permissions mask */
	umask((mode_t)oldmask);

}


/*========================================================================
 *	Name:			XPWriteAddresses()
 *	Prototype:	void XPWriteAddresses()
 *					
 *	Description:
 *	Write the Address list back to disk.
 *
 *	Input Arguments:
 *		char	*filename		name of file to write to
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *		
 *========================================================================*/
void 
XPWriteAddresses(
	char	*filename
)
{
	FILE					*fp;
	char					tmpfile[MAXPATHLEN];
	mode_t				oldmask;
	PilotAddr_T			*current_link;
	PilotCategory_T	*category_link;
	char					cmd[MAXPATHLEN];
	char					magicmark[0xffff];

	/*
	 * Open a temporary file - make sure it's only readable by the user.
	 */
   sprintf(tmpfile,"%s/xp3.%d", app_res.note_dir, (int)getpid());
	oldmask = umask((mode_t)022);

   /*
    * Open the first temporary file for writing
    */
	if ((fp = fopen(tmpfile, "w")) == NULL) 
	{
		/* Not likely, but you never know */
		fprintf(stderr, "xnotes: ");
		perror(tmpfile);
		return;
	}

	/* Write out the magic string */
	sprintf(magicmark, MAGICMARK, "addresses");
	fprintf(fp, "%s\n", magicmark);

	/*
	 * Write out the categories first. 
	 */
	category_link = Pilot_Addr_Categories;
	while (category_link != NULL)
	{
		fprintf(fp, "%s\n", category_link->name);
		category_link = category_link->next;
	}


	/*
	 * Start with the first link.
	 */
	current_link = Pilot_Address;

	/*
	 * For each link, write out the entry.
	 */
	while (current_link != NULL)
	{
		fprintf(fp, "%sCategory: %s\n", PREFIX, current_link->category);
		fprintf(fp, "%sRecordID: %d\n", PREFIX, current_link->recordid);
		fprintf(fp, "%sListPhone: %s\n", PREFIX, current_link->listPhone);
		fprintf(fp, "%sAttribute: %d\n", PREFIX, current_link->attributes);

		/* Name, Company */
		if ( strlen(current_link->last) > 1 )
			fprintf(fp, "%sLast: %s\n", PREFIX, current_link->last);
		else
			fprintf(fp, "%sLast: (null)\n", PREFIX);
		if ( strlen(current_link->first) > 1 )
			fprintf(fp, "%sFirst: %s\n", PREFIX, current_link->first);
		else
			fprintf(fp, "%sFirst: (null)\n", PREFIX);
		if ( strlen(current_link->title) > 1 )
			fprintf(fp, "%sTitle: %s\n", PREFIX, current_link->title);
		else
			fprintf(fp, "%sTitle: (null)\n", PREFIX);
		if ( strlen(current_link->company) > 0 )
			fprintf(fp, "%sCompany: %s\n", PREFIX, current_link->company);
		else
			fprintf(fp, "%sCompany: (null)\n", PREFIX);

		/* Phone, Fax, email, etc. */
		if ( strlen(current_link->phone1) > 1 )
			fprintf(fp, "%s%s: %s\n", PREFIX, 
				current_link->phoneLabel1, current_link->phone1);
		else
			fprintf(fp, "%s%s: (null)\n", PREFIX, current_link->phoneLabel1);

		if ( strlen(current_link->phone2) > 1 )
			fprintf(fp, "%s%s: %s\n", PREFIX, 
				current_link->phoneLabel2, current_link->phone2);
		else
			fprintf(fp, "%s%s: (null)\n", PREFIX, current_link->phoneLabel2);

		if ( strlen(current_link->phone3) > 1 )
			fprintf(fp, "%s%s: %s\n", PREFIX, 
				current_link->phoneLabel3, current_link->phone3);
		else
			fprintf(fp, "%s%s: (null)\n", PREFIX, current_link->phoneLabel3);

		if ( strlen(current_link->phone4) > 1 )
			fprintf(fp, "%s%s: %s\n", PREFIX, 
				current_link->phoneLabel4, current_link->phone4);
		else
			fprintf(fp, "%s%s: (null)\n", PREFIX, current_link->phoneLabel4);

		if ( strlen(current_link->phone5) > 1 )
			fprintf(fp, "%s%s: %s\n", PREFIX, 
				current_link->phoneLabel5, current_link->phone5);
		else
			fprintf(fp, "%s%s: (null)\n", PREFIX, current_link->phoneLabel5);

		/* Address, City, State, etc. */
		if ( strlen(current_link->address) > 1 )
			fprintf(fp, "%sAddress: %s\n", PREFIX, current_link->address);
		else
			fprintf(fp, "%sAddress: (null)\n", PREFIX);
		if ( strlen(current_link->city) > 1 )
			fprintf(fp, "%sCity: %s\n", PREFIX, current_link->city);
		else
			fprintf(fp, "%sCity: (null)\n", PREFIX);
		if ( strlen(current_link->state) > 1 )
			fprintf(fp, "%sState: %s\n", PREFIX, current_link->state);
		else
			fprintf(fp, "%sState: (null)\n", PREFIX);
		if ( strlen(current_link->zip) > 1 )
			fprintf(fp, "%sZipCode: %s\n", PREFIX, current_link->zip);
		else
			fprintf(fp, "%sZipCode: (null)\n", PREFIX);
		if ( strlen(current_link->country) > 1 )
			fprintf(fp, "%sCountry: %s\n", PREFIX, current_link->country);
		else
			fprintf(fp, "%sCountry: (null)\n", PREFIX);
		
		/* Custom Fields */
		if ( strlen(current_link->custom1) > 1 )
			fprintf(fp, "%sCustom1: %s\n", PREFIX, current_link->custom1);
		else
			fprintf(fp, "%sCustom1: (null)\n", PREFIX);
		if ( strlen(current_link->custom2) > 1 )
			fprintf(fp, "%sCustom2: %s\n", PREFIX, current_link->custom2);
		else
			fprintf(fp, "%sCustom2: (null)\n", PREFIX);
		if ( strlen(current_link->custom3) > 1 )
			fprintf(fp, "%sCustom3: %s\n", PREFIX, current_link->custom3);
		else
			fprintf(fp, "%sCustom3: (null)\n", PREFIX);
		if ( strlen(current_link->custom4) > 1 )
			fprintf(fp, "%sCustom4: %s\n", PREFIX, current_link->custom4);
		else
			fprintf(fp, "%sCustom4: (null)\n", PREFIX);
		
		/* Note field */
		if(strlen(current_link->note) > 1)
			fprintf(fp, "%sNote: %s\n", PREFIX, current_link->note );
		else
		{
			fprintf(fp, "%sNote: (null)\n", PREFIX);

			/* Add a blank line */
			fprintf(fp, "\n");
		}

		current_link = current_link->next;
	}

	/*
	 * Close the file and copy it over the original.
	 * This is to help prevent partial overwrite during system crashes.
	 * It's not foolproof, but what file operation is atomic?
	 */
	fclose(fp);
	if ( rename(tmpfile, filename) != 0 )
	{
		sprintf(cmd, "xnotes: %s", tmpfile);
		perror(cmd);
		sprintf(cmd,"Couldn't rename %s\nto %s.\n"
			"Save of Address List failed.",
			tmpfile, filename);
		XPErrPopUp(cmd);
		umask((mode_t)oldmask);
		return;
	}

	/* Reset the file permissions mask */
	umask((mode_t)oldmask);

}


/*========================================================================
 *	Name:			XPAddAddress()
 *	Prototype:	void XPAddAddress()
 *					
 *	Description:
 *	Add a new link to the Address link list, then sort things again.
 *
 *	Notes:
 *		
 *========================================================================*/
void 
XPAddAddress(
	PilotAddr_T		*ptr
)
{
	PilotAddr_T		*current_link;

	if ( Pilot_Address == NULL )
		Pilot_Address = ptr;
	else
	{
		current_link = Pilot_Address;
		while (current_link->next != NULL)
			current_link = current_link->next;
		current_link->next = ptr;
	}

	SortAddresses();
	SortCategoryList(&Pilot_Addr_Categories);
}


/*========================================================================
 *	Name:			XPPilotParseDBList
 *	Prototype:	static void XPPIlotParseDBList(char *filename)
 *					
 *	Description:
 * Parse the list of DB names from a temporary file.
 *
 *	Notes:
 * This is hear just to keep things consistant.
 *========================================================================*/
void
XPPilotParseDBList(
	char	*filename,
	char	**text_data
)
{
	FILE			*fd;
	int			filelen;
	char			*file_data;
	gboolean		ready;
	int			ptr, text_ptr;

	fd = fopen(filename, "r");
	fseek(fd, 0, SEEK_END);
	filelen = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	file_data = (char*)XPSafeAlloc(filelen+1);
	*text_data = (char*)XPSafeAlloc(filelen+1);
	fread(file_data, filelen, 1, fd);

	/* Clean up the text */
	ptr = (int)file_data;
	text_ptr = (int)*text_data;
	ready = FALSE;
	while ( ptr < (int)((int)file_data + filelen) )
	{
		if (! ready)
		{
			if (strncmp((char *)ptr, "'", 1) == 0 )
				ready = TRUE;
			ptr++;
		}
		else
		{
			if (strncmp((char *)ptr, "'", 1) != 0 )
			{
				if (strncmp((char *)ptr, "List done.", 10) != 0 )
				{
					strncpy((char *)text_ptr, (char *)ptr, 1);
					text_ptr++;
					ptr++;
				}
				else
					ptr += 10;
			}
			else
				ptr++;
		}
	}
	free(file_data);
	fclose(fd);
}


/*========================================================================
 *	Name:			XPPilotParseAppointment
 *	Prototype:	void XPPilotParseAppointment(char *filename)
 *					
 *	Description:
 *	Parses the Appointment List flat file into a link list.
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *		XP_E_FILE_EXIST		file to parse doesn't exist or is not a text file
 *		XP_E_SUCCESS			All went well.
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *	This version uses the new pseudo-XML format.  
 *		
 *========================================================================*/
int 
XPPilotParseAppointment(
	char	*filename
)
{
#ifdef DEBUG
	char				*fname="XPPiloParseAppointment()";
#endif

	PilotAppt_T		*pentry;
	PilotAppt_T		*current_link=NULL;

	struct stat		stat_buf;
	FILE				*fd;
	char				*file_text;
	int				filelen;
	int				cPtr, endPtr;
	int				count;
	char				*name, *value;
	int				id;
	char				*idchr;
	char				*chr;
	int				j;


	DBGEnter();
	DBGPrintf(DBG_INFO,("pilot db to look for: %s\n", filename));

	/*
	 * Verify file exists.  Return if it doesn't.
	 */
	if ( stat(filename, &stat_buf) != 0 )
		return XP_E_FILE_EXIST;

	if ( ! ( stat_buf.st_mode & S_IFREG ) )
		return XP_E_FILE_EXIST;

	/*
	 * File is ok.  Open it and read in the file.
	 */
	fd = fopen(filename, "r");

	/*
	 * Free up any old link lists.
	 */
	XPFreeApptList(&Pilot_Appt);

	/*
	 * Allocate a new link for an address record.
	 */
	pentry = (PilotAppt_T *)XPSafeAlloc(sizeof(PilotAppt_T));
	memset(pentry,0,sizeof(PilotAppt_T));

	Pilot_Appt = pentry;
	current_link = pentry;

	/* Read the file into memory - it's faster this way. */
	fseek(fd, 0, SEEK_END);
	filelen = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	file_text = (char*)malloc(filelen+1);
	if (file_text == NULL) 
	{
		/* This should never happen. */
		perror("malloc()");
		exit(1);
	}
   fread(file_text, filelen, 1, fd);
	fclose(fd);

	/*
	 * Initialize pointers, counters used to parse input text.
	 */
	cPtr = (int)file_text;
	count = 0;

	/*
	 * Move past the Magic Mark
	 */
	name=strtok((char *)cPtr, "\n");
	cPtr = cPtr + strlen(name) + 1;

	/*
	 * Move past the marker that separates the category
	 * list from the records.  (Note:  Appt's don't have categories 
	 * currently, but this could change someday.)
	 */
	while( strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 )
	{
		name=strtok((char *)cPtr, "\n");
		cPtr = cPtr + strlen(name) + 1;
	}

	/*
	 * Grab each records data.
	 * 
	 * Record format:
	 * RecordID
	 * Attributes
	 * int Event
	 * struct tm EventStart
	 * struct tm EventEnd
	 * int Alarm
	 * int Advance
	 * int AdvanceUnits
	 * int RepeatType
	 * int RepeatForever
	 * struct tm RepeatEnd
	 * int RepeatFrequency
	 * int RepeatDay
	 * int RepeatDays[7] (int)
	 * int RepeatWeekstart
	 * int Exceptions
	 * struct tm Exception (multiple)
	 * char *Description
	 * char *Note
	 */
	while( cPtr < (int)((int)file_text + filelen) )
	{

		/* Move past the record prefix */
		cPtr = cPtr + strlen(PREFIX);

		/*
		 * Get the current name, value pair.
		 */
		name=strtok((char *)cPtr, ":");
		cPtr = cPtr + strlen(name) + 2;
		value = (char *)cPtr;
	
		while ( 
			(strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 ) && 
			((int)cPtr < (int)((int)file_text + filelen) )
			)
			{
				cPtr++;
			}

		/* 
		 * Remove trailing newline.
		 */
		endPtr = cPtr - 1;

		/*
		 * Null terminate the string 
		 */
		strncpy((char *)endPtr, "\0", 1);


		/* Grab the record id and convert to an integer */
		if ( strncmp(name, "RecordID", strlen("RecordID")) == 0 )
		{
			current_link->recordid = (unsigned long)atoi(value);
		}
		else

		/* Grab the record attributes and convert to an integer */
		if ( strncmp(name, "Attributes", strlen("Attributes")) == 0 )
		{
			current_link->attributes = atoi(value);
		}
		else

		/* Grab the category name and convert it to an category id */
		if ( strncmp(name, "Category", strlen("Category")) == 0 )
		{
			/* No categories for appointments. */
		}
		else

		/* Grab the event type (ie is it timeless?). */
		if ( strncmp(name, "Event", strlen("Event")) == 0 )
		{
			current_link->event = atoi(value);
		}
		else

		/* Grab the Begin date */
		if ( strncmp(name, "Begin", strlen("Begin")) == 0 )
		{
			strptime(value, "%a %b %d %T %Y", &current_link->begin);
		}
		else

		/* Grab the End date */
		if ( strncmp(name, "End", strlen("End")) == 0 )
		{
			strptime(value, "%a %b %d %T %Y", &current_link->end);
		}
		else

		/* Grab the Alarm */
		if ( strncmp(name, "Alarm", strlen("Alarm")) == 0 )
		{
			current_link->alarm = atoi(value);
		}
		else

		/* Grab the AdvanceUnits (do before "advance"!) */
		if ( strncmp(name, "AdvanceUnits", strlen("AdvanceUnits")) == 0 )
		{
			current_link->advanceUnits = atoi(value);
		}
		else

		/* Grab the Advance */
		if ( strncmp(name, "Advance", strlen("Advance")) == 0 )
		{
			current_link->advance = atoi(value);
		}
		else

		/* Grab the RepeatType */
		if ( strncmp(name, "RepeatType", strlen("RepeatType")) == 0 )
		{
			current_link->repeatType = atoi(value);
		}
		else

		/* Grab the RepeatForever */
		if ( strncmp(name, "RepeatForever", strlen("RepeatForever")) == 0 )
		{
			current_link->repeatForever = atoi(value);
		}
		else

		/* Grab the RepeatEnd date */
		if ( strncmp(name, "RepeatEnd", strlen("RepeatEnd")) == 0 )
		{
			if ( current_link->repeatType == 0 ) 
			{
				bzero(&current_link->repeatEnd, sizeof(struct tm));
			}
			else
			{
				strptime(value, "%a %b %d %T %Y", &current_link->repeatEnd);
			}
		}
		else

		/* Grab the RepeatFrequency */
		if ( strncmp(name, "RepeatFrequency", strlen("RepeatFrequency")) == 0 )
		{
			current_link->repeatFrequency = atoi(value);
		}
		else

		/* Grab the RepeatDay */
		if ( strncmp(name, "RepeatDay", strlen("RepeatDay")) == 0 )
		{
			current_link->repeatDay = atoi(value);
		}
		else

		/* Grab the RepeateWeekStart */
		if ( strncmp(name, "RepeastWeekstart", strlen("RepeatWeekstart")) == 0 )
		{
			current_link->repeatWeekstart = atoi(value);
		}
		else

		/* Grab the multiple Exceptions */
		if ( strncmp(name, "ExceptionIDS", strlen("ExceptionIDS")) == 0 )
		{
			if ( current_link->exception != NULL )
			{
				id = atoi((char *)(name+12));
				strptime(value, "%a %b %d %T %Y", &current_link->exception[id]);
			}
		}
		else

		/* Grab the Exceptions */
		if ( strncmp(name, "Exceptions", strlen("Exceptions")) == 0 )
		{
			current_link->exceptions = atoi(value);
			if ( current_link->exceptions > 0 )
				current_link->exception = 
					(struct tm *)malloc(sizeof(struct tm)*current_link->exceptions);
			else
				current_link->exception = NULL;
		}
		else


		/* Grab the RepeatDates */
		if ( strncmp(name, "RepeatDates", strlen("RepeatDates")) == 0 )
		{
			idchr = (char *)malloc(strlen(value)+1);
			strcpy(idchr, value);
			bzero((char *)(idchr+strlen(idchr)-1), 1);
			id = atoi(strtok(idchr, ":"));
			for (j=0; j<7; j++)
			{
				current_link->repeatDays[j] = id;
				chr = strtok(NULL, ":");
				if ( chr ) id = atoi(chr);
			}
			free(idchr);
		}
		else

		/* Grab the Description */
		if ( strncmp(name, "Description", strlen("Description")) == 0 )
		{
			current_link->description = (char *)malloc(strlen(value)+1);
			sprintf(current_link->description, "%s", value);
		}
		else

		/* Grab the Note */
		if ( strncmp(name, "Note", strlen("Note")) == 0 )
		{
			current_link->note = (char *)malloc(strlen(value)+1);
			sprintf(current_link->note, "%s", value);
		}
		else

		if ( strncmp(name, "RecordEND", strlen("RecordEND")) == 0 )
		{
			/* DEBUG */
#if 0
			printf(
				"Recordid:        %d\n"
	 			"Attributes:      %d\n"
	 			"Event:           %d\n"
	 			"Begin:           %s\n"
	 			"End:             %s\n"
	 			"Alarm:           %d\n"
	 			"Advance:         %d\n"
	 			"AdvanceUnits:    %d\n"
	 			"repeatType:      %d\n"
	 			"repeatForever:   %d\n"
	 			"repeatEnd:       %s\n"
	 			"repeatFrequency: %d\n"
	 			"repeatDay:       %d\n"
	 			"repeatWeekstart: %d\n"
	 			"exceptions:      %d\n",
				current_link->recordid,
				current_link->attributes,
				current_link->event,
				asctime(&current_link->begin),
				asctime(&current_link->end),
				current_link->alarm,
				current_link->advance,
				current_link->advanceUnits,
				current_link->repeatType,
				current_link->repeatForever,
				asctime(&current_link->repeatEnd),
				current_link->repeatFrequency,
				current_link->repeatDay,
				current_link->repeatWeekstart,
				current_link->exceptions);
#endif

			for (j=0; j<current_link->exceptions; j++)
			{
				printf("Exception %d: %s\n", j, asctime(&current_link->exception[j]));
			}

			if ( cPtr < (int)((int)file_text + filelen) )
			{
				pentry = (PilotAppt_T *)XPSafeAlloc(sizeof(PilotAppt_T));
				memset(pentry,0,sizeof(PilotAppt_T));
				current_link->next = pentry;
				current_link = pentry;
			}
		}
	}

	return ( XP_E_SUCCESS );
}


/*========================================================================
 *	Name:			XPPilotWriteAppointment
 *	Prototype:	void XPPilotWriteAppointment(char *filename)
 *					
 *	Description:
 *	Write the Appointment List flat file from the link list.  
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *	This version uses the new pseudo-XML format.  
 *========================================================================*/
void 
XPPilotWriteAppointment(
	char	*filename
)
{
#ifdef DEBUG
	char				*fname="XPPiloWriteAppointment()";
#endif

	PilotAppt_T		*current_link=NULL;
	char				magicmark[0xffff];
	char				tmpfile[MAXPATHLEN];
	mode_t			oldmask;
	char				cmd[MAXPATHLEN];
	FILE				*fp;
	int				j;

	DBGEnter();

	/* Do nothing if there are no entries. */
	if ( Pilot_Appt == NULL )
	{
		DBGPrintf(DBG_INFO, ("No appointments to write to %s\n", filename));
		printf("No appointments to write to file.\n");

		/* If there are no appointments, make sure the file doesn't exist. */
		unlink(filename);
		return;
	}

	/*
	 * Open a temporary file - make sure it's only readable by the user.
	 */
   sprintf(tmpfile,"%s/xp3.%d", app_res.pilot_dir, (int)getpid());
	oldmask = umask((mode_t)022);
	DBGPrintf(DBG_INFO, ("Writing to %s (then %s)\n", tmpfile, filename));

   /*
    * Open the first temporary file for writing
    */
	if ((fp = fopen(tmpfile, "w")) == NULL) 
	{
		/* Not likely, but you never know */
		fprintf(stderr, "xnotes: ");
		perror(tmpfile);
		return;
	}

	/* Write out the magic string */
	sprintf(magicmark, MAGICMARK, "datebook");
	fprintf(fp, "%s\n", magicmark);

	/*
	 * Start with the first link.
	 */
	current_link = Pilot_Appt;

	/*
	 * For each link, write out the entry.
	 */
	while (current_link != NULL)
	{
		/* Write the record to the file. */
		fprintf(fp, "%sRecordID: %d\n", PREFIX, (int)current_link->recordid);
		fprintf(fp, "%sAttributes: %d\n", PREFIX, (int)current_link->attributes);
		fprintf(fp, "%sEvent: %d\n", PREFIX, current_link->event);
		fprintf(fp, "%sBegin: %s", PREFIX, asctime(&current_link->begin));
		fprintf(fp, "%sEnd: %s", PREFIX, asctime(&current_link->end));
		fprintf(fp, "%sAlarm: %d\n", PREFIX, current_link->alarm);
		fprintf(fp, "%sAdvance: %d\n", PREFIX, current_link->advance);
		fprintf(fp, "%sAdvanceUnits: %d\n", PREFIX, current_link->advanceUnits);
		fprintf(fp, "%sRepeatType: %d\n", PREFIX, current_link->repeatType);
		fprintf(fp, "%sRepeatForever: %d\n", PREFIX, current_link->repeatForever);
		if ( current_link->repeatType == 0 ) bzero(&current_link->repeatEnd, sizeof(struct tm));
		fprintf(fp, "%sRepeatEnd: %s", PREFIX, asctime(&current_link->repeatEnd));
		fprintf(fp, "%sRepeatFrequency: %d\n", PREFIX, current_link->repeatFrequency);
		fprintf(fp, "%sRepeatDay: %d\n", PREFIX, current_link->repeatDay);
		fprintf(fp, "%sRepeatDates: ", PREFIX);
		for (j=0; j<7; j++)
			fprintf(fp, "%d:", current_link->repeatDays[j]);
		fprintf(fp, "\n");
		fprintf(fp, "%sRepeatWeekstart: %d\n", PREFIX, current_link->repeatWeekstart);
		fprintf(fp, "%sExceptions: %d\n", PREFIX, current_link->exceptions);
		for (j=0; j<current_link->exceptions; j++)
			fprintf(fp, "%sExceptionID%d: %s", 
				PREFIX, j, asctime(&current_link->exception[j]));
		if ( current_link->description )
			fprintf(fp, "%sDescription: %s\n", PREFIX, current_link->description);
		else
			fprintf(fp, "%sDescription: \n", PREFIX);
		if ( current_link->note )
			fprintf(fp, "%sNote: %s\n", PREFIX, current_link->note);
		else
			fprintf(fp, "%sNote: \n", PREFIX);
		fprintf(fp, "%sRecordEND:\n", PREFIX);

		current_link = current_link->next;
	}

	/*
	 * Close the file and copy it over the original.
	 * This is to help prevent partial overwrite during system crashes.
	 * It's not foolproof, but what file operation is atomic?
	 */
	fclose(fp);
	if ( rename(tmpfile, filename) != 0 )
	{
		sprintf(cmd, "xnotes: %s", tmpfile);
		perror(cmd);
		sprintf(cmd,"Couldn't rename %s\nto %s.\n"
			"Save of Appointment List failed.",
			tmpfile, filename);
		XPErrPopUp(cmd);
		umask((mode_t)oldmask);
		return;
	}

	/* Reset the file permissions mask */
	umask((mode_t)oldmask);
}


/*========================================================================
 *	Name:			XPFreeApptList
 *	Prototype:	void XPFreeApptList(char *filename)
 *					
 *	Description:
 *	Write the Appointment List flat file from the link list.  
 *
 *	Input Arguments:
 *		char	*filename		name of file with addresses in it.
 *	Output Arguments:
 *	Return Values:
 *	Global Variables:
 *	External Routines:
 *	Method:
 *	Restrictions:
 *	Notes:
 *	This version uses the new pseudo-XML format.  
 *========================================================================*/
void 
XPFreeApptList()
{
#ifdef DEBUG
	char				*fname="XPFreeApptList()";
#endif

	PilotAppt_T	*current_link=NULL;
	PilotAppt_T	*next_link=NULL;

	DBGEnter();

	/*
	 * Start with the first link.
	 */
	current_link = Pilot_Appt;
	while ( current_link != NULL )
	{
		next_link = current_link->next;
		if ( current_link->description ) free(current_link->description);
		if ( current_link->note ) free(current_link->note);
		if ( current_link->exceptions ) free((char *)current_link->exceptions);
		current_link = next_link;
	}
}
